class Permissions(object):

    _JAVA_PKG = "com.genestack.file."

    FILE_ACCESS = _JAVA_PKG + "access"
    FILE_READ_CONTENT = _JAVA_PKG + "readContent"
    FILE_WRITE = _JAVA_PKG + "write"
    FILE_CLONE_DATA = _JAVA_PKG + "cloneData"
